import uuid
from datetime import datetime
from django.db import models
from django.core.validators import RegexValidator
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.contrib.contenttypes.fields import GenericForeignKey
from django.core.cache import cache

from django_form_generator.common.models import BaseModel
from django_form_generator.common.helpers import FileFieldHelper
from django_form_generator.common.utils import (
    APICall,
    evaluate_data,
)

from django_form_generator import const
from django_form_generator.settings import form_generator_settings as fg_settings


CONTENT_TYPE_MODELS_LIMIT = models.Q(
    app_label="django_form_generator", model="field"
) | models.Q(app_label="django_form_generator", model="value")


class FieldCategory(BaseModel):
    title = models.CharField(_("Title"), max_length=200, unique=True)
    is_active = models.BooleanField(_("Is Active"), default=True)
    parent = models.ForeignKey(
        "self",
        verbose_name=_("Parent Category"),
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        limit_choices_to={"is_active": True},
        related_name="children",
    )
    weight = models.PositiveIntegerField(_("Weight"))

    class Meta:
        verbose_name = _("FieldCategory")
        verbose_name_plural = _("FieldCategories")
        ordering = ("weight",)
        indexes = [
            models.Index(fields=("title",), name="f_g_%(class)s_title"),
            models.Index(fields=("is_active",), name="f_g_%(class)s_active"),
            models.Index(fields=("weight",), name="f_g_%(class)s_weight"),
        ]

    def __str__(self) -> str:
        return self.title

    def get_fields(self):
        return self.fields.filter(is_active=True)  # type: ignore


class Form(BaseModel):
    title = models.CharField(_("Title"), max_length=200)
    slug = models.SlugField(_("Slug"), unique=True, allow_unicode=True)
    status = models.CharField(
        _("Status"),
        max_length=50,
        choices=const.FormStatus.choices,
        default=const.FormStatus.DRAFT,
    )
    submit_text = models.CharField(
        _("Submit Text"), max_length=100, default=_("Submit")
    )
    redirect_url = models.URLField(
        _("Redirect URL"), max_length=200, blank=True, null=True
    )
    success_message = models.CharField(
        _("Success Message"), max_length=255, blank=True, null=True
    )
    theme = models.CharField(
        _("Theme"),
        max_length=300,
        blank=True,
        null=True,
        help_text=_(
            "If choose dynamic_fields.html the order of fields depends on position of Field."
        ),
    )
    direction = models.CharField(
        _("Direction"),
        max_length=3,
        choices=const.FormDirection.choices,
        default=const.FormDirection.LTR,
    )
    limit_to = models.PositiveIntegerField(
        _("Limit Submiting Form"), blank=True, null=True
    )
    valid_from = models.DateTimeField(_("Valid From"), blank=True, null=True)
    valid_to = models.DateTimeField(_("Valid To"), blank=True, null=True)
    fields = models.ManyToManyField(
        "django_form_generator.Field",
        verbose_name=_("Fields"),
        through="django_form_generator.FormFieldThrough",
        related_name="forms",
    )
    apis = models.ManyToManyField(
        "django_form_generator.FormAPIManager",
        verbose_name=_("Apis"),
        through="django_form_generator.FormAPIThrough",
        related_name="forms",
    )
    is_editable = models.BooleanField(_("Is Editable"), default=False)

    objects = fg_settings.FORM_MANAGER()  # type: ignore

    class Meta:
        verbose_name = _("Form")
        verbose_name_plural = _("Forms")

        indexes = [
            models.Index(fields=("title",), name="f_g_%(class)s_title"),
            models.Index(fields=("slug",), name="f_g_%(class)s_slug"),
            models.Index(fields=("status",), name="f_g_%(class)s_status"),
        ]

    def __str__(self) -> str:
        return self.title

    def get_absolute_url(self):
        return reverse("django_form_generator:form_detail", kwargs={"pk": self.pk})

    def __call_apis(
        self, execute_time: const.FormAPIManagerExecuteTime, response_data: dict
    ):
        cache_key = f"FormAPIs_{self.pk}_{execute_time}_{response_data['request'].session.session_key}"
        cached_response = cache.get(cache_key)
        if cached_response:
            return cached_response

        responses = []
        for api in self.apis.filter(is_active=True, execute_time=execute_time):
            api: FormAPIManager
            response = APICall(
                api.method.lower(),
                api.url,
                api.body,
                response_data,
                headers=api.headers,
            )
            status_code, result = response.get_result()
            responses.append((api, status_code, result))
        cache.set(cache_key, responses)
        return responses

    def call_post_apis(self, response_data: dict):
        return self.__call_apis(
            const.FormAPIManagerExecuteTime.POST_LOAD, response_data
        )

    def render_post_apis(self, response_data: dict):
        data = []
        results = self.call_post_apis(response_data)
        for api, _, result in results:
            res = evaluate_data(api.response, result)
            data.append(res)
        return data

    def call_pre_apis(self, response_data: dict):
        return self.__call_apis(const.FormAPIManagerExecuteTime.PRE_LOAD, response_data)

    def render_pre_apis(self, response_data: dict):
        data = []
        results = self.call_pre_apis(response_data)
        for api, _, result in results:
            res = evaluate_data(api.response, result)
            data.append((api.pk, res))
        return data

    def get_fields(self, extra: dict | None = None):
        conds = {"is_active": True}
        if extra:
            conds.update(extra)
        return self.fields.filter(**conds).order_by(
            models.F("form_field_through__category__weight").asc(
                nulls_last=True),
            "form_field_through__weight",
        )


class FormFieldThrough(models.Model):
    form = models.ForeignKey(
        "django_form_generator.Form",
        verbose_name=_("Form"),
        on_delete=models.CASCADE,
        related_name="form_field_through",
    )
    field = models.ForeignKey(
        "django_form_generator.Field",
        verbose_name=_("Field"),
        on_delete=models.CASCADE,
        related_name="form_field_through",
    )
    position = models.CharField(
        _("Position"),
        max_length=100,
        choices=const.FieldPosition.choices,
        default=const.FieldPosition.BREAK,
    )
    category = models.ForeignKey(
        "django_form_generator.FieldCategory",
        verbose_name=_("Category"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        limit_choices_to={"is_active": True},
        related_name="form_field_through",
    )
    weight = models.PositiveIntegerField(_("Weight"))

    class Meta:
        ordering = ("weight",)
        indexes = [
            models.Index(fields=("weight",), name="f_g_%(class)s_weight"),
        ]

    def __str__(self) -> str:
        return f"{self.form.title} | {self.field.name}"


class Field(BaseModel):
    label = models.CharField(_("Label"), max_length=100)
    name = models.CharField(_("Name"), max_length=100, blank=True, unique=True)
    genre = models.CharField(
        _("Genre"), max_length=80, choices=const.FieldGenre.choices
    )
    is_required = models.BooleanField(_("Is Required"), default=True)
    placeholder = models.CharField(
        _("Placeholder"), max_length=100, blank=True, null=True
    )
    default = models.CharField(
        _("Default Value"), max_length=255, blank=True, null=True
    )
    help_text = models.CharField(
        _("Help Text"), max_length=200, blank=True, null=True)
    regex_pattern = models.CharField(
        _("Regex Pattern"), max_length=500, blank=True, null=True
    )
    error_message = models.CharField(
        _("Error Message"), max_length=200, blank=True, null=True
    )
    values = models.ManyToManyField(
        "django_form_generator.Value",
        through="django_form_generator.FieldValueThrough",
        verbose_name=_("Values"),
        related_name="fields",
        help_text=_(
            "Only for multi value fields like Dropdown, Radio, Checkbox, etc..."
        ),
    )
    is_active = models.BooleanField(_("Is Active"))
    # file
    file_types = models.CharField(
        _("Allow File Types"),
        max_length=100,
        blank=True,
        null=True,
        help_text=_("example: jpg,png"),
    )
    file_size = models.IntegerField(
        _("File Size"),
        blank=True,
        null=True,
        help_text=_("Default value will be set to 50 MB"),
    )
    # depends
    content_type = models.ForeignKey(
        "contenttypes.ContentType",
        verbose_name=_("Depends On Object"),
        on_delete=models.SET_NULL,
        related_name="fields",
        limit_choices_to=CONTENT_TYPE_MODELS_LIMIT,
        blank=True,
        null=True,
    )
    object_id = models.BigIntegerField(
        _("Depends on Object ID"), blank=True, null=True)
    content_object = GenericForeignKey()

    class Meta:
        verbose_name = _("Field")
        verbose_name_plural = _("Fields")
        indexes = [
            models.Index(fields=("label",), name="f_g_%(class)s_label"),
            models.Index(fields=("name",), name="f_g_%(class)s_name"),
            models.Index(fields=("genre",), name="f_g_%(class)s_genre"),
        ]

    def __str__(self) -> str:
        return self.label

    def build_field_attrs(self, extra_attrs: dict | None = None):
        attrs = {
            "required": self.is_required,
            "help_text": self.help_text,
            "label": self.label,
        }

        if self.error_message:
            attrs.update({"error_messages": {"Invalid": self.error_message}})

        if self.regex_pattern is not None:
            attrs.update(
                {
                    "validators": [
                        RegexValidator(self.regex_pattern,
                                       self.error_message or None)
                    ]
                }
            )

        if self.default is not None:
            attrs.update({"initial": self.default})

        if extra_attrs:
            attrs.update(extra_attrs)
        return attrs

    def build_widget_attrs(self, form, extra_attrs: dict | None = None):
        form_field_through = self.form_field_through.filter(form_id=form.id).last()  # type: ignore
        attrs = {"instance_id": self.pk,
                 "position": form_field_through.position}

        if self.genre in const.FieldGenre.selectable_fields():
            attrs.update({"onchange": "onElementChange(this)"})
        else:
            attrs.update({"onkeypress": "onElementChange(this)"})

        if self.content_object is not None:
            attrs.update(
                {
                    "parent_object_id": self.object_id,
                    "parent_content_type": self.content_type.model,
                    "disabled": True
                }
            )

        if form_field_through.category is not None:
            attrs.update({"category": form_field_through.category.title})

        if self.placeholder is not None:
            attrs.update({"placeholder": self.placeholder})

        if extra_attrs:
            attrs.update(extra_attrs)
        return attrs

    def get_choices(self):
        return self.values.filter(is_active=True).order_by("field_values__weight")

    def get_file_types(self):
        return self.file_types.replace(" ", ",").replace(", ", ",").split(",")


class FieldValueThrough(models.Model):
    field = models.ForeignKey(
        "django_form_generator.Field",
        verbose_name=_("Field"),
        on_delete=models.CASCADE,
        related_name="field_values",
    )
    value = models.ForeignKey(
        "django_form_generator.Value",
        verbose_name=_("Value"),
        on_delete=models.CASCADE,
        related_name="field_values",
    )
    weight = models.PositiveIntegerField(_("Weight"))

    class Meta:
        ordering = ("weight",)
        indexes = [models.Index(
            fields=("weight",), name="f_g_%(class)s_weight")]

    def __str__(self) -> str:
        return f"{self.field.name} | {self.value.name}"


class Value(BaseModel):
    name = models.CharField(_("Name"), max_length=100)
    is_active = models.BooleanField(_("Is Active"), default=True)

    class Meta:
        verbose_name = _("Value")
        verbose_name_plural = _("Values")
        indexes = [
            models.Index(fields=("name",), name="f_g_%(class)s_name"),
        ]

    def __str__(self) -> str:
        return self.name


class FormAPIThrough(models.Model):
    form = models.ForeignKey(
        "django_form_generator.Form",
        verbose_name=_("Form"),
        on_delete=models.CASCADE,
        related_name="form_apis",
    )
    api = models.ForeignKey(
        "django_form_generator.FormAPIManager",
        verbose_name=_("API"),
        on_delete=models.CASCADE,
        related_name="form_apis",
    )
    weight = models.PositiveIntegerField(_("Weight"))

    class Meta:
        ordering = ("weight",)
        indexes = [models.Index(
            fields=("weight",), name="f_g_%(class)s_weight")]

    def __str__(self) -> str:
        return f"{self.form.title} | {self.api.title}"


class FormAPIManager(BaseModel):
    title = models.CharField(_("Title"), max_length=200)
    url = models.URLField(_("URL"), max_length=200)
    headers = models.JSONField(_("Headers"), blank=True, null=True)
    method = models.CharField(
        _("Method"), max_length=50, choices=const.FormAPIManagerMethod.choices
    )
    body = models.TextField(_("Body"), blank=True, null=True)
    execute_time = models.CharField(
        _("Execute Time"),
        max_length=50,
        choices=const.FormAPIManagerExecuteTime.choices,
    )
    response = models.TextField(_("Response"), blank=True, null=True)
    is_active = models.BooleanField(_("Is Active"))

    class Meta:
        verbose_name = _("FormAPIManager")
        verbose_name_plural = _("FormAPIManagers")
        indexes = [
            models.Index(fields=("title",), name="f_g_%(class)s_title"),
            models.Index(fields=("method",), name="f_g_%(class)s_method"),
            models.Index(fields=("is_active",), name="f_g_%(class)s_active"),
        ]

    def __str__(self) -> str:
        return self.title


class FormResponseBase(BaseModel):
    unique_id = models.UUIDField(
        _("Unique ID"), unique=True, default=uuid.uuid4)
    form = models.ForeignKey(
        "django_form_generator.Form",
        verbose_name=_("Form"),
        on_delete=models.PROTECT,
        related_name="%(class)s_responses",
    )
    data = models.JSONField(_("Data"))
    api_response = models.JSONField(_("Api Respons"), blank=True, null=True)

    class Meta:
        abstract = True

    def _str_(self):
        return self.form.title

    def save(self, *args, **kwargs):
        if self.unique_id is None:
            self.unique_id = uuid.uuid4()
        return super().save(*args, **kwargs)

    @property
    def pure_data(self):
        result = {}
        for d in self.data:
            result[d["name"]] = d["value"]  # * {field_name: User input data}
        return result

    def get_data(self):
        result = []
        for data in self.data:
            data_ = data.copy()
            if (
                data["genre"] == const.FieldGenre.UPLOAD_FILE.value
                and data["value"] is not None
            ):
                data_["value"] = FileFieldHelper(
                    data["value"]["url"], data["value"]["directory"]
                )
                result.append(data_)
            elif data_["value"] and data["genre"] in (const.FieldGenre.DATETIME, const.FieldGenre.DATE):
                data_["value"] = datetime.fromisoformat(data_["value"])
                result.append(data_)
            else:
                result.append(data_)
        return result

    @classmethod
    def _generate_api_result(cls, results):
        data = []
        for api, status_code, result in results:
            data.append(
                {
                    "api": api.id,
                    "url": api.url,
                    "method": api.method,
                    "body": api.body,
                    "response_status_code": status_code,
                    "result": result,
                }
            )
        return data or None

    @classmethod
    def _generate_data(cls, form, form_data, instance=None):
        data = []
        request = form_data["request"]
        for field in form.get_fields():
            field_value = form_data.get(field.name, None)
            if field.genre in (const.FieldGenre.DATE,
                               const.FieldGenre.TIME,
                               const.FieldGenre.DATETIME):
                field_value = str(field_value)
            if field.genre == const.FieldGenre.CAPTCHA:
                continue
            category_title = None
            category = field.form_field_through.filter(
                form_id=form.pk).last().category
            if category:
                category_title = category.title
            if (
                field_value is not None
                and not isinstance(field_value, dict)
                and field.genre == const.FieldGenre.UPLOAD_FILE
            ):
                try:
                    if instance is not None:
                        instance_directory = instance.pure_data[field.name]["directory"]
                except:
                    instance_directory = None
                form_data[field.name] = FileFieldHelper.upload_file(
                    request._current_scheme_host, field_value, instance_directory
                )
            if field_value and  isinstance(field_value, str):
                field_value = int(field_value) if field_value.isdigit() else field_value
            elif isinstance(field_value, list):
                field_value = [int(val) if isinstance(val, str) and val.isdigit() else val for val in field_value]

            new_data = {
                "id": field.id,
                "name": field.name,
                "label": field.label,
                "genre": field.genre,
                "category": category_title,
                "value": field_value,
                "depends_on": None
            }
            if field.content_object:
                new_data["depends_on"] = {"id": field.object_id, 
                                            "type": field.content_type.model,
                                            "value": next((form_data.get(i.name, None) for i in form.get_fields() if i.name == field.content_object.name), None)}
            data.append(
                new_data
            )
        return data or None


class FormResponse(FormResponseBase):
    user_ip = models.GenericIPAddressField(
        _("IP Address"), blank=True, null=True)

    class Meta:
        verbose_name = _("Form Response")
        verbose_name_plural = _("Form Responses")
        indexes = [
            models.Index(fields=("user_ip",), name="f_g_%(class)s_user"),
            models.Index(fields=("unique_id",),
                         name="f_g_%(class)s_unique_id"),
        ]

    @classmethod
    def save_response(cls, form, data, user_ip=None, update_form_response_id=None):
        if update_form_response_id is None:
            api_response = []
            pre_result = form.call_pre_apis(data)
            post_result = form.call_post_apis(data)
            pre = cls._generate_api_result(pre_result)
            post = cls._generate_api_result(post_result)
            if pre:
                api_response.append(pre)
            if post:
                api_response.append(post)
            response = FormResponse.objects.create(
                data=cls._generate_data(form, data),
                user_ip=user_ip,
                api_response=api_response or None,
                form=form,
            )
        else:
            response = FormResponse.objects.get(id=update_form_response_id)
            response.data = cls._generate_data(form, data, response)
            response.save()
        return response


def save_form_response(
    form: Form,
    form_data: dict,
    user_ip: str | None = None,
    update_form_response_id: int | None = None,
):
    # * You can access `request` from form_data
    FormResponse.save_response(
        form, form_data, user_ip, update_form_response_id)



from django.urls import path
from django_form_generator.views import FormGeneratorView, FormResponseView


urlpatterns = [
    path('form/<int:pk>/', FormGeneratorView.as_view(), name="form_detail"),
    path('form-response/<uuid:unique_id>/', FormResponseView.as_view(), name="form_response"),
]



import uuid
from django.db import models
from django.contrib import admin
from django.utils.text import slugify
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.db.transaction import atomic

from django_form_generator import const
from django_form_generator.common.admins import FormFilter
from django_form_generator.forms import FieldForm, FormAdminForm, FormResponseFilterForm
from django_form_generator.models import (
    FieldCategory,
    Form,
    Field,
    FormFieldThrough,
    Value,
    FormResponse,
    FormAPIThrough,
    FieldValueThrough,
    FormAPIManager,
)

# Register your models here.




class FormResponseFilter(FormFilter):

    template = "django_form_generator/filters/form_response_filter.html"
    form_class = FormResponseFilterForm

    def _create_related_fields_lookup(self, q, fields, fields_text, current_index, current_field_id, related_field_id, index=None):
        related =  Field.objects.filter(object_id=related_field_id, content_type__model='field').last()
        if related and str(related.pk) in fields:
            depended_field = Field.objects.filter(object_id=current_field_id, content_type__model='field').last()
            if depended_field:
                try:
                    depended_field_index = current_index + 1
                    fields_text[depended_field_index]
                except IndexError:
                    depended_field_index = current_index 
                dp_value = int(fields_text[depended_field_index]) if fields_text[depended_field_index].isdigit() else fields_text[depended_field_index]
                if index:
                    if isinstance(dp_value, int):
                        q.add(models.Q(**{f"{self.field_path}__{index}__value": dp_value}), models.Q.AND)
                    else:
                        q.add(models.Q(**{f"{self.field_path}__{index}__value__regex": dp_value}), models.Q.AND)
                else:
                    q.add(models.Q(**{f"{self.field_path}__contains": [{"id": depended_field.pk, "value": dp_value}]}), models.Q.AND)
        return q

    def get_lookups(self):
        lookup = models.Q()

        fields_text = self.request.GET.getlist(f'{self.field.name}-text')
        fields_id = self.request.GET.getlist(f'{self.field.name}-field')
        fields_operand = self.request.GET.getlist(f'{self.field.name}-operand')
        form_id = int(self.request.GET.get('form_id', '0'))
        if form_id:
            form = Form.objects.prefetch_related('fields').get(id=form_id)
            form_fields_list = list(form.get_fields().values_list('id', flat=True))

            for i in range(len(fields_text)):
                field_id = int(fields_id[i])
                field_value = int(fields_text[i]) if fields_text[i].isdigit() else fields_text[i]
                current_operand = getattr(models.Q, fields_operand[i], models.Q.AND)
                if isinstance(field_value, int):
                    lookup_content = "{0}__{1}__value"
                else:
                    lookup_content = "{0}__{1}__value__regex"
                try:
                    index = form_fields_list.index(field_id)
                    inner_lookup = models.Q(**{lookup_content.format(self.field_path, index): field_value})
                    self._create_related_fields_lookup(inner_lookup, fields_id, fields_text, i, field_id, field_id, index+1)
                    lookup.add(inner_lookup, current_operand)
                except ValueError:
                    # searching on multi_fields
                    inner_lookup = models.Q()
                    current_field = Field.objects.get(id=field_id)
                    multi_fields = Field.objects.filter(name__startswith=current_field.name).order_by('name')
                    if multi_fields.count() > 1:
                        for field_ in multi_fields[1:]:
                            index = form_fields_list.index(field_.pk)
                            related_lookup = models.Q(**{lookup_content: field_value})
                            self._create_related_fields_lookup(related_lookup, fields_id, fields_text, i, field_.pk, multi_fields[0].pk, index+1)
                            inner_lookup.add(related_lookup, models.Q.OR)
                        lookup.add(inner_lookup, current_operand)
        return lookup


    def form_lookups(self):
        name = self.field.name
        return (
            ("%s-field" % name, "%s__contains" % name),
            ("%s-text" % name, "%s__contains" % name),
            ("%s-operand" % name, "%s__contains" % name),
        )


class FormFieldThroughInlineAdmin(admin.TabularInline):
    model = FormFieldThrough
    extra = 1
    raw_id_fields = ("field",)


class FormAPIThroughInlineAdmin(admin.TabularInline):
    model = FormAPIThrough
    extra = 1
    raw_id_fields = ("api",)


@admin.register(Form)
class FormAdmin(admin.ModelAdmin):
    list_display = ['id', 'title', 'status', 'get_theme', 'created_at', 'updated_at']
    list_display_links = ['id', 'title']
    list_editable = ['status']
    list_filter = ['status', 'created_at']
    search_fields = ['title', 'slug']
    readonly_fields = ['id', 'created_at', 'updated_at']
    inlines = [FormFieldThroughInlineAdmin, FormAPIThroughInlineAdmin]
    form = FormAdminForm
    actions = ("clone_action",)

    fieldsets = (
        (None, {
            'fields': ('title', 'slug', 'status', 'submit_text', 'redirect_url', 'success_message', 'id', 'created_at', 'updated_at')
        }),
        ('Style', {
            'classes': ('wide',),
            'fields': ('theme', 'direction'),
        }),
        ('Limitations', {
            'classes': ('wide',),
            'fields': ('limit_to', 'valid_from', 'valid_to', 'is_editable'),
        }),
    )

    @admin.display(description="Theme")
    def get_theme(self, obj):
        return const.FormTheme(obj.theme).label

    @atomic
    def clone_action(self, request, queryset):
        for obj in queryset:
            f_through = obj.form_field_through.all().values('field_id', 'position', 'category_id', 'weight')
            a_through = obj.form_apis.all().values('api_id', 'weight')
            for field in obj._meta.local_fields:
                if field.unique:
                    if field.name in ("pk", "id"):
                        setattr(obj, field.name, None)

            setattr(obj, "slug", uuid.uuid4())
            setattr(obj, 'title', obj.title + ' (Copy)')
            setattr(obj, "status", const.FormStatus.DRAFT)
            obj.save()

            ff_through = [FormFieldThrough(form_id=obj.id, **obj_) for obj_ in f_through]
            FormFieldThrough.objects.bulk_create(ff_through, ignore_conflicts=True)
            fa_through = [FormAPIThrough(form_id=obj.id, **obj_) for obj_ in a_through]
            FormAPIThrough.objects.bulk_create(fa_through)

class FieldValueThroughInlineAdmin(admin.TabularInline):
    model = FieldValueThrough
    extra = 1
    raw_id_fields = ("value",)


@admin.register(Field)
class FieldAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'genre', 'is_active', 'created_at', 'updated_at']
    list_display_links = ['id', 'name']
    list_editable = ['is_active']
    list_filter = ['is_active', 'created_at', 'genre']
    search_fields = ['label', 'name', 'forms__title']
    inlines = [FieldValueThroughInlineAdmin]
    readonly_fields = ['id', 'name', 'created_at', 'updated_at']
    form = FieldForm

    fieldsets = (
        (None, {
            'fields': ('label', 'genre', 'is_required', 'placeholder', 'default', 'help_text', 'is_active', 'regex_pattern', 'error_message', 'id', 'created_at', 'updated_at')
        }),
        ('File', {
            'classes': ('wide',),
            'fields': ('file_types', 'file_size'),
        }),
        ('Dpendency', {
            'classes': ('wide',),
            'fields': ('content_type', 'object_id'),
        }),
    )

    def save_form(self, request, form, change):
        form.instance.name = slugify(form.instance.label, True).replace("-", "_")
        return super().save_form(request, form, change)


@admin.register(FormResponse)
class FormResponseAdmin(admin.ModelAdmin):
    list_display = ["id", "get_form_title", "user_ip", "show_response"]
    list_display_links = ["id", "get_form_title"]
    list_filter = [('data', FormResponseFilter)]
    search_fields = ['form__title', 'form__slug']
    readonly_fields = ['id', 'unique_id', 'created_at', 'updated_at']

    @admin.display(description="Form Title")
    def get_form_title(self, obj):
        return obj.form.title

    @admin.display(description="Response")
    def show_response(self, obj):
        return mark_safe(
            f"<a href='{reverse('django_form_generator:form_response', args=(obj.unique_id, ))}'>Response</a>"
        )

@admin.register(FormAPIManager)
class FormAPIManagerAdmin(admin.ModelAdmin):
    list_display = ["id", "title", "method", "execute_time", "is_active", "created_at", "updated_at"]
    list_display_links = ["id", "title"]
    list_filter = ['is_active', 'created_at', 'execute_time', 'method']
    list_editable = ['is_active']
    search_fields = ['title', 'forms__title', 'forms__slug']
    readonly_fields = ['id', 'created_at', 'updated_at']


@admin.register(FieldCategory)
class FieldCategoryAdmin(admin.ModelAdmin):
    list_display = ["id", "title", "parent", "weight", "is_active", "created_at", "updated_at"]
    list_display_links = ["id", "title"]
    list_filter = ['is_active', 'created_at', ('parent', admin.filters.EmptyFieldListFilter)]
    list_editable = ['is_active']
    search_fields = ['title']
    readonly_fields = ['id', 'created_at', 'updated_at']


@admin.register(Value)
class ValueAdmin(admin.ModelAdmin):
    list_display = ["id", "name", "is_active", "created_at", "updated_at"]
    list_display_links = ["id", "name"]
    list_filter = ['is_active', 'created_at']
    list_editable = ['is_active']
    search_fields = ['name', 'fields__name', 'fields__label']
    readonly_fields = ['id', 'created_at', 'updated_at']




import urllib
from django.contrib import admin


class FormFilter(admin.FieldListFilter):

    initial = {}
    form_class = None

    def __init__(self, field, request, params, model, model_admin, field_path):
        super().__init__(field, request, params, model, model_admin, field_path)
        self.request = request
        expected = self.expected_parameters()
        others = {k: v for k, v in request.GET.items() if k not in expected}
        self.other_query_string = urllib.parse.urlencode(others)
        self.form = self.get_form(request)
        if self.form:
            self.form.is_valid()

    def form_lookups(self):
        raise NotImplementedError(
            "subclasses of FormFieldFilter must provide a form_lookups() method"
        )

    def expected_parameters(self):
        return [item[0] for item in self.form_lookups()]

    def get_initial(self):
        return self.initial.copy()

    def get_form_kwargs(self, request):
        return {
            "prefix": self.field.name,
            "initial": self.get_initial(),
            "data": request.GET or None,
        }

    def get_form(self, request):
        return self.form_class(**self.get_form_kwargs(request))

    def get_lookups(self):
        lookups = {k.split("_")[1]: v for k, v in self.form_lookups()}
        data = self.form.cleaned_data if self.form.is_bound else {}
        return {v: data[k] for k, v in lookups.items() if data.get(k)}

    def queryset(self, request, queryset):
        lookup = self.get_lookups()
        if isinstance(lookup, dict):
            return queryset.filter(**lookup)
        return queryset.filter(lookup)

    def choices(self, changelist):
        return ()




from django import forms
from django.core.validators import FileExtensionValidator
from django.utils.translation import gettext as _

from captcha.fields import ReCaptchaField
from captcha.widgets import ReCaptchaV2Checkbox
from tempus_dominus.widgets import DatePicker, TimePicker, DateTimePicker

from django_form_generator.settings import form_generator_settings as fg_settings
from django_form_generator.common.utils import FileSizeValidator
from django_form_generator.models import Field, Form, Value
from django_form_generator import const


class CustomeSelectFormField(forms.Select):
    option_inherits_attrs = True


class FormGeneratorBaseForm(forms.Form):
    def __init__(self, form, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = form
        self.template_name_p = getattr(self.instance, "theme", self.template_name_p)
        self._initial_fields()
        
    def _initial_fields(self):
        for field in self.instance.get_fields():
            method = f"prepare_{field.genre}"
            if hasattr(self, method):
                self.fields[field.name] = getattr(self, method)(field)
                self._handel_required_fields(field, self.fields[field.name])

    def _handel_required_fields(self, field, form_field):
        if self.data and field.content_object:
            if isinstance(field.content_object, Value):
                field_name = self.instance.get_fields(extra={"id__in":field.content_object.fields.values_list('id', flat=True)}).last().name
                parent_data = self.data.get(field_name, '')
                if str(field.object_id) not in parent_data:
                    form_field.required = False
                else:
                    form_field.widget.attrs.update({'disabled': False})
            else:
                parent_data = self.data.get(field.content_object.name, '')
                if len(parent_data) <= 0:
                    form_field.required = False
                else:
                    form_field.widget.attrs.update({'disabled': False})

    def prepare_text_input(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.TextInput(attrs=widget_attrs)}
        )
        return forms.CharField(**field_attrs)

    def prepare_text_area(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.Textarea(attrs=widget_attrs)}
        )
        return forms.CharField(**field_attrs)

    def prepare_number(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.NumberInput(attrs=widget_attrs)}
        )
        return forms.IntegerField(**field_attrs)

    def prepare_date(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": DatePicker(attrs=widget_attrs)}
        )
        return forms.DateField(**field_attrs)

    def prepare_time(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": TimePicker(attrs=widget_attrs)}
        )
        return forms.TimeField(**field_attrs)

    def prepare_datetime(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": DateTimePicker(attrs=widget_attrs)}
        )
        return forms.DateTimeField(**field_attrs)

    def prepare_email(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.EmailInput(attrs=widget_attrs)}
        )
        return forms.EmailField(**field_attrs)

    def prepare_password(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.PasswordInput(attrs=widget_attrs)}
        )
        return forms.CharField(**field_attrs)

    def prepare_checkbox(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.CheckboxInput(attrs=widget_attrs)}
        )
        return forms.BooleanField(**field_attrs)

    def prepare_dropdown(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "value"}
        )
        choices = field.get_choices().values_list("id", "name")
        field_attrs: dict = field.build_field_attrs(
            {
                "widget": CustomeSelectFormField(attrs=widget_attrs),
                "choices": tuple(choices),
            }
        )
        return forms.ChoiceField(**field_attrs)

    def prepare_multi_checkbox(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "value"}
        )
        choices = field.get_choices().values_list("id", "name")
        field_attrs: dict = field.build_field_attrs(
            {
                "widget": forms.CheckboxSelectMultiple(attrs=widget_attrs),
                "choices": tuple(choices),
            }
        )
        return forms.MultipleChoiceField(**field_attrs)

    def prepare_radio(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "value"}
        )
        choices = field.get_choices().values_list("id", "name")
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.RadioSelect(attrs=widget_attrs), "choices": tuple(choices)}
        )
        return forms.ChoiceField(**field_attrs)

    def prepare_hidden(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": forms.HiddenInput(attrs=widget_attrs)}
        )
        return forms.CharField(**field_attrs)

    def prepare_captcha(self, field: Field):
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {"widget": ReCaptchaV2Checkbox(attrs=widget_attrs)}
        )
        return ReCaptchaField(**field_attrs)

    def prepare_upload_file(self, field: Field):
        message = (
            _("The file size is more than limit (limited size: %s bytes)")
            % field.file_size
        )
        widget_attrs: dict = field.build_widget_attrs(
            self.instance, {"multiple": True, "content_type": "field"}
        )
        field_attrs: dict = field.build_field_attrs(
            {
                "widget": forms.ClearableFileInput(attrs=widget_attrs),
                "validators": [
                    FileSizeValidator(field.file_size, message),
                    FileExtensionValidator(field.get_file_types()),
                ],
            }
        )
        return forms.FileField(**field_attrs)


class FormGeneratorForm(FormGeneratorBaseForm):
    def __init__(self, form, request, user_ip, *args, **kwargs):
        self.user_ip = user_ip
        self.request = request
        super().__init__(form, *args, **kwargs)

    def save(self):
        form_data = self.cleaned_data.copy()
        form_data.setdefault("request", self.request)
        save_module = fg_settings.FORM_RESPONSE_SAVE  
        save_module(self.instance, form_data, self.user_ip)# type: ignore


class FormGeneratorResponseForm(FormGeneratorBaseForm):
    def __init__(self, form, request, form_response, *args, **kwargs):
        self.request = request
        self.form_response = form_response
        super().__init__(form, *args, **kwargs)

    def _initial_fields(self):
        form_response_data = self.form_response.get_data()
        for i, field in enumerate(self.instance.get_fields()):
            field_name = field.name
            method = f"prepare_{field.genre}"
            if hasattr(self, method):
                self.fields[field_name] = getattr(self, method)(field)
                if not self.instance.is_editable:
                    self.fields[field_name].widget.attrs.update({"disabled": True})
                try:
                    initial_value = form_response_data[i].get(
                        "value", None
                    )
                    self.fields[field_name].initial = initial_value
                    if initial_value:
                        try:
                            del self.fields[field_name].widget.attrs["disabled"]
                        except KeyError:
                            pass
                except IndexError:
                    pass
                
                self._handel_required_fields(field, self.fields[field.name])

    def save(self):
        save_module = fg_settings.FORM_RESPONSE_SAVE
        form_data = self.form_response.pure_data
        for changed_data in self.changed_data:
            form_data[changed_data] = self.cleaned_data[changed_data]

        form_data.setdefault("request", self.request)
        save_module(self.instance, form_data, update_form_response_id=self.form_response.id)# type: ignore
        


class FieldForm(forms.ModelForm):
    class Meta:
        model = Field
        fields = "__all__"

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data["genre"] == const.FieldGenre.UPLOAD_FILE:
            if not cleaned_data["file_types"]:
                raise forms.ValidationError(
                    _("You should define FileTypes for upload file genre")
                )
        return cleaned_data

    def save(self, commit=True):
        if self.cleaned_data["genre"] == const.FieldGenre.UPLOAD_FILE and (
            not self.cleaned_data["file_size"] or self.cleaned_data["file_size"] <= 0
        ):
            self.instance.file_size = fg_settings.MAX_UPLOAD_FILE_SIZE
        return super().save(commit)


class FormAdminForm(forms.ModelForm):
    theme = forms.ChoiceField(choices=fg_settings.FORM_THEME_CHOICES.choices)  # type: ignore

    class Meta:
        model = Form
        fields = "__all__"



class FormResponseFilterForm(forms.Form):
    operand = forms.ChoiceField(choices=(('OR', 'OR'), ('AND', 'AND')),  required=False)
    field = forms.ModelChoiceField(Field.objects.all(), required=False)
    text = forms.CharField(max_length=120, required=False, help_text=_("if want to search on DropDown field or Radio or ... enter ID of it from Value table"))


import os
import pathlib
from typing import Any
from django.conf import settings
from django_form_generator.common.utils import upload_file_handler

class FileFieldHelper:

    def __init__(self, url_path, directory_path):
        self.url_path = url_path
        self.directory_path = directory_path

    @property
    def url(self):
        return self.url_path

    @property
    def name(self):
        name = os.path.basename(self.directory_path)
        return name

    @classmethod
    def upload_file(cls, host: str, data: bool|Any, instance_directory: str|None=None):
        if isinstance(data, bool) and not data:
            if instance_directory is not None:
                pathlib.Path(instance_directory).unlink()
            return None
        directory = upload_file_handler(data)
        parts = pathlib.Path(directory).parts
        return {
            "directory": directory,
            "url": host + settings.MEDIA_URL + "/".join(parts[-2:]),
        }


from django.db import models
from django.utils.translation import gettext_lazy as _


class BaseModel(models.Model):
    created_at = models.DateTimeField(_("Created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("Updated at"), auto_now=True)

    class Meta:
        abstract = True


import requests
import uuid
import os
from pathlib import Path
from django.conf import settings
from django.template import Template, Context
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.core.validators import BaseValidator

from django_form_generator.const import FormAPIManagerMethod
from django_form_generator.settings import form_generator_settings as fg_settings


FILE_UPLOAD_DIRECTORY = os.path.join(settings.MEDIA_ROOT, 'django_form_generator')

try:
    Path(FILE_UPLOAD_DIRECTORY).mkdir(exist_ok=True)
except Exception as e:
    print(e)

def upload_file_handler(f):
    new_filename = uuid.uuid4()
    _, ext = os.path.splitext(f.name)
    dest = FILE_UPLOAD_DIRECTORY + f'/{new_filename}{ext}'
    with open(dest, 'wb+') as destination:
        for chunk in f.chunks():
            destination.write(chunk)
    return dest


def get_client_ip(request):
    """get the client IP address"""
    remote_address = request.META.get('REMOTE_ADDR')
    ip = remote_address
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        proxies = x_forwarded_for.split(',')
        while (len(proxies) > 0):
            proxies.pop(0)
        if len(proxies) > 0:
            ip = proxies[0]
    return ip


def evaluate_data(data: str, replace_with: dict|list) -> str | list:
    """replace data that wrapped in a pattern like: {{sample}} with provided dictionary

    Args:
        data (str): 'hello {{first_name}} {{last_name}}'
        replace_with (dict): {'first_name': 'john', 'last_name', 'doe'}

    Returns:
        str: 'hello john doe'
    """
    if isinstance(replace_with, list):
        l_data = []
        for item in replace_with:
            l_data.append(evaluate_data(data, item))
        return l_data
    elif isinstance(replace_with, dict):
        if fg_settings.FORM_EVALUATIONS['form_data'] in data:
            replace_with['form_data'] = mark_safe(replace_with)
        template = Template(data)
        context = Context(replace_with)
        data = template.render(context)
        return data
    else:
        return replace_with


class APICall:

    def __init__(self, method, url, body: str|None=None, data_response: dict|list|None=None, **kwargs):
        request = getattr(requests, method)
        if data_response is not None:
            url = evaluate_data(url, data_response)
        if method == FormAPIManagerMethod.GET:
            response = request(url, **kwargs)
        else:
            if data_response is not None:
                body = evaluate_data(body, data_response) #type: ignore
            response = request(url, data=body, **kwargs)
        result = response.json()
        self.status_code: int = response.status_code
        self.result: dict = result

    def get_result(self) -> tuple[int, dict]:
        return  self.status_code, self.result


class FileSizeValidator(BaseValidator):

    def compare(self, file_, limit_value):
        return file_.size > limit_value


from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _

class FormStatus(TextChoices):
    DRAFT = 'draft', _('Draft')
    PENDING = 'pending', _('Pending')
    PUBLISH = 'publish', _('Publish')
    SUSPEND = 'suspend', _('Suspend')

class FieldGenre(TextChoices):
    TEXT_INPUT = 'text_input', _('Text input')
    TEXT_AREA = 'text_area', _('Text area')
    NUMBER = 'number', _('Number')
    DROPDOWN = 'dropdown', _('Dropdown')
    DATE = 'date', _('Date')
    TIME = 'time', _('Time')
    DATETIME = 'datetime', _('Datetime')
    EMAIL = 'email', _('E-Mail')
    PASSWORD = 'password', _('Password')
    CHECKBOX = 'checkbox', _('Checkbox')
    MULTI_CHECKBOX = 'multi_checkbox', _('Multi checkbox')
    RADIO = 'radio', _('Radio')
    HIDDEN = 'hidden', _('Hidden')
    CAPTCHA = 'captcha', _('Captcha')
    UPLOAD_FILE = 'upload_file', _('Upload file')

    @classmethod
    def selectable_fields(cls):
        return [cls.DROPDOWN, cls.CHECKBOX, cls.RADIO, cls.MULTI_CHECKBOX]

class FormAPIManagerMethod(TextChoices):
    GET = 'get', _('GET')
    POST = 'post', _('POST')
    PUT = 'put', _('PUT')
    PATCH = 'patch', _('PATCH')
    DELETE = 'delete', _('DELETE')
    OPTION = 'option', _('OPTION')

class FormAPIManagerExecuteTime(TextChoices):
    PRE_LOAD = 'pre_load', _('Pre load')
    POST_LOAD = 'post_load', _('Post load')

class FieldPosition(TextChoices):
    INLINE = 'inline', _('In-line')
    INORDER = 'inorder', _('In-Order')
    BREAK = 'break', _('Break')

class FormDirection(TextChoices):
    LTR = 'ltr', _('LTR')
    RTL = 'rtl', _('RTL')

class FormTheme(TextChoices):
    INLINE = 'django_form_generator/fields/inline_fields.html', _('In-line Theme')
    INORDER = 'django_form_generator/fields/inorder_fields.html', _('In-order Theme')
    DYNAMIC = 'django_form_generator/fields/dynamic_fields.html', _('Dynamic Theme')


from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

DEFAULTS = {
    'FORM_RESPONSE_SAVE': 'django_form_generator.models.save_form_response',
    'FORM_EVALUATIONS': {'form_data': '{{form_data}}'},
    'MAX_UPLOAD_FILE_SIZE': 5242880,
    'FORM_GENERATOR_FORM': 'django_form_generator.forms.FormGeneratorForm',
    'FORM_RESPONSE_FORM': 'django_form_generator.forms.FormGeneratorResponseForm',
    'FORM_THEME_CHOICES': 'django_form_generator.const.FormTheme',
    'FORM_MANAGER': 'django_form_generator.managers.FormManager',
    'FORM_GENERATOR_RESPONSE_MODEL': 'django_form_generator.models.FormResponse',
}


IMPORT_STRINGS = [
    'FORM_RESPONSE_SAVE',
    'FORM_GENERATOR_FORM',
    'FORM_RESPONSE_FORM',
    'FORM_THEME_CHOICES',
    'FORM_MANAGER',
    'FORM_GENERATOR_RESPONSE_MODEL',
]

def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for FORM_GENERATOR setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)

class FormGeneratorSettings:
    def __init__(self, defaults=None, import_strings=None):
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'DJANGO_FORM_GENERATOR', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Form generator setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


form_generator_settings = FormGeneratorSettings(DEFAULTS, IMPORT_STRINGS)


def reload_form_generator_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'DJANGO_FORM_GENERATOR':
        form_generator_settings.reload()


setting_changed.connect(reload_form_generator_settings)